/*
openIRC
Copyright (C) 2006  target0 (target0@geeknode.org)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <fcntl.h>
#include <sys/poll.h>
#include <setjmp.h>
#include <signal.h>
#include <stdarg.h>
#include "openirc.h"

char *ircMsg[] = {
                NULL,       /* IRC_MSG_RAW */
                "PRIVMSG",  /* IRC_MSG_PRIV */
                "NOTICE",   /* IRC_MSG_NOTICE */
                "QUIT",     /* IRC_MSG_QUIT */
                "KICK",     /* IRC_MSG_KICK */
                "INVITE",   /* IRC_MSG_INVITE */
                "MODE",     /* IRC_MSG_MODE */
                "JOIN",     /* IRC_MSG_JOIN */
                "PART",     /* IRC_MSG_PART */
                "TOPIC",    /* IRC_MSG_TOPIC */
                };

jmp_buf timeout_jump;

void timeout()
{
    longjmp(timeout_jump, 1);
}

ircClient *ircGetClient()
{
    ircClient *client = (ircClient *)calloc(1, sizeof(ircClient));
    client->indata.nextline = client->indata.chunkbuf;
    client->indata.chunkbufentry = client->indata.chunkbuf;
    client->outdata.pfd.events = POLLOUT;
    client->outdata.pfd.revents = 0;
    client->indata.pfd.events = POLLIN | POLLPRI;
    client->indata.pfd.revents = 0;

    return client;
}

void ircFreeClient(ircClient *client)
{
    while (client->callbacks != NULL)
        ircRemoveCallback(client, client->callbacks);
    free(client);
}

int ircConnect(ircClient *client)
{
    struct addrinfo hints, *res = NULL;
    int n;

    hints.ai_flags = 0;
    hints.ai_family = PF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;
    hints.ai_addrlen = 0;
    hints.ai_addr = NULL;
    hints.ai_canonname = NULL;
    hints.ai_next = NULL;

    char port[6];
    sprintf(port,"%d", client->port);

    n = getaddrinfo(client->hostname, port, &hints, &res);
    if (n != 0) {
        fprintf(stderr, "ircError: getaddrinfo: %s\n", gai_strerror(n));
        if (n == EAI_SYSTEM) perror("getaddrinfo");
        return ERR_FAILED;
    }

    client->sock = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if (client->sock < 0) {
        perror("socket");
        return ERR_FAILED;
    }

    signal(SIGALRM, timeout);
    alarm(client->timeout);
    if (setjmp(timeout_jump) == 1) {
        close(client->sock);
        return ERR_TIMEOUT;
    }

    if (connect(client->sock, res->ai_addr, res->ai_addrlen) == -1) {
        perror("connect");
        alarm(0);
        signal(SIGALRM, SIG_DFL);
        close(client->sock);
        return ERR_FAILED;
    }

    alarm(0);
    signal(SIGALRM, SIG_DFL);
    freeaddrinfo(res);

    client->indata.pfd.fd = client->outdata.pfd.fd = client->sock;
    fcntl(client->sock, F_SETFL, O_NONBLOCK);

    return ERR_SUCCESS;
}

int ircSend (ircClient *client, char *data, ...)
{
    char tmp[1024];
    char buf[1024];
    va_list val;
    int len;
    char *outptr;
    va_start(val, data);
    vsnprintf(buf, 1023, data, val);
    va_end(val);

    snprintf(tmp, 1024, "%s\r\n",buf);

    len = strlen(tmp);
    if (len > SENDQSIZE || len + client->outdata.wbytes >= SENDQSIZE)
        return ERR_MAXSENDQ;

    outptr = client->outdata.outbuf + client->outdata.wbytes;

    strncpy(outptr, tmp, SENDQSIZE - client->outdata.wbytes - 1);
    client->outdata.wbytes += len;

    return ERR_SUCCESS;
}

static void irc__flush_sendq (ircClient *client)
{
    if (client->outdata.wbytes == 0)
        return;

    send(client->sock, client->outdata.outbuf, strlen(client->outdata.outbuf), 0);

    client->outdata.outbuf[0] = 0;
    client->outdata.wbytes = 0;
}

void ircHandShake (ircClient *client)
{
    ircSend(client, "NICK %s", client->nick);
    ircSend(client, "USER %s . . :%s", client->ident, client->realname);
}

void ircDisconnect (ircClient *client)
{
    close(client->sock);
}

int ircRead(ircClient *client)
{
    int oldbytes;
    int readbytes;
    char *dest;

    if (*client->indata.nextline != '\0') {
        oldbytes = strlen(client->indata.nextline);
        dest = client->indata.chunkbuf;
        while (*client->indata.nextline)
            *dest++ = *client->indata.nextline++;
        client->indata.chunkbufentry = client->indata.chunkbuf + oldbytes;
    } else {
        oldbytes = 0;
        client->indata.chunkbufentry = client->indata.chunkbuf;
    }

    readbytes = read(client->sock, client->indata.chunkbufentry, (CHUNKSIZE - oldbytes) - 10);
    if (readbytes == 0 || readbytes == -1)
        return 0;

    *(client->indata.chunkbufentry + readbytes) = '\0';
    client->indata.nextline = client->indata.chunkbuf;

    return 1;
}

int ircGetLine (ircClient *client)
{
    client->indata.currentline = client->indata.nextline;
    while (*client->indata.nextline != '\n' && *client->indata.nextline != '\r' &&
            *client->indata.nextline != '\0')
        client->indata.nextline++;

    if (*client->indata.nextline == '\0') {
        client->indata.nextline = client->indata.currentline;
        return 0;
    }

    if (*client->indata.nextline == '\n' || *client->indata.nextline == '\r') {
        *client->indata.nextline = '\0';
        client->indata.nextline++;
    }

    if (*client->indata.nextline == '\n' || *client->indata.nextline == '\r') {
        *client->indata.nextline = '\0'; 
        client->indata.nextline++; 
    }

    return 1;
}

char *SeperateWord(char *sentence)
{
    if (sentence == NULL)
        return NULL;

    while (*sentence != ' ' && *sentence != '\0' && *sentence != '\n')
        sentence++;

    if (*sentence == '\0' || *sentence == '\n') {
        *sentence = '\0';
        return (NULL);
    }

    if (*sentence == ' ') {
        *sentence = '\0';
        sentence++;
    }

    return (StripBlanks(sentence));
}

char *StripBlanks(char *paddedstring)
{
    if (paddedstring == NULL)
        return NULL;

    while (*paddedstring == ' ')
        paddedstring++;

     return paddedstring;
}

int ircLoop (ircClient *client)
{
    int retval;

    if (client->outdata.wbytes > 0) {
        retval = poll(&client->outdata.pfd, 1, 1000);
        if (retval > 0 && (client->outdata.pfd.revents & POLLOUT))
            irc__flush_sendq(client);
    }
    retval = poll(&client->indata.pfd, 1, 1000);
    if (retval > 0 && (client->indata.pfd.revents & (POLLIN | POLLPRI))) {
        if (!ircRead(client))
            return ERR_FAILED;
        while (ircGetLine(client))
            ircParse(client);
        return ERR_SUCCESS;
    }

    return ERR_NODATA;
}

ircCallback *ircAddCallback(ircClient *client, int id, char *command, void (*func)())
{
    ircCallback *callback;
    callback = (ircCallback *)calloc(1, sizeof(ircCallback));
    callback->id = id;
    if (id == IRC_MSG_RAW)
        strncpy(callback->command, command, 63);
    callback->func = func;

    if (client->callbacks != NULL) {
        client->callbacks->prev = callback;
        callback->next = client->callbacks;
    }

    client->callbacks = callback;

    return callback;
}

ircCallback *ircFindCallback(ircClient *client, char *command)
{
    ircCallback *tmp;
    for (tmp = client->callbacks; tmp; tmp = tmp->next) {
        if (tmp->id == IRC_MSG_RAW) {
            if (!strcasecmp(tmp->command, command))
                return tmp;
        } else {
            if (!strcasecmp(ircMsg[tmp->id], command))
                return tmp;
        }
    }

    return NULL;
}

void ircRemoveCallback(ircClient *client, ircCallback *callback)
{
    if (client->callbacks == callback)
        client->callbacks = NULL;
    else if (callback->prev == NULL) {
        client->callbacks = callback->next;
        client->callbacks->prev = NULL;
     } else if (callback->next == NULL)
        callback->prev->next = NULL;

    free(callback);
};

void ircParse(ircClient *client)
{
    char *sender, *command, *tail;
    ircCallback *callback;

    sender = StripBlanks(client->indata.currentline);
    command = SeperateWord(sender);
    tail = SeperateWord(command);

    if (!sender || *sender == '\0' || !command || *command == '\0')
        return;

    if (!strcasecmp(sender, "PING")) {
        ircSend(client, "PONG %s", command);
        return;
    }

    if ((callback = ircFindCallback(client, command)) != NULL)
        callback->func(client, sender, tail);
}

ircUser *ircParseUser (char *sender)
{
    char tmp[512];
    strncpy(tmp, sender, 511);

    char *ptr = tmp;
    if (*ptr++ != ':')
        return NULL;

    char *nick, *ident, *host;
    nick = ptr;
    if ((ptr = strchr(ptr, '!')) == NULL)
        return NULL;

    *ptr++ = '\0';
    ident = ptr;
    if ((ptr = strchr(ptr, '@')) == NULL)
        return NULL;

    *ptr++ = '\0';
    host = ptr;

    ircUser *user = (ircUser *)calloc(1, sizeof(ircUser));
    strncpy(user->nick, nick, NICKLEN);
    strncpy(user->ident, ident, NICKLEN);
    strncpy(user->host, host, HOSTLEN);

    return user;
}

ircMsgData *ircParseData (char *data)
{
    char tmp[1024];
    strncpy(tmp, data, 1023);

    char *ptr = tmp;
    char *target;

    ircMsgData *msgdata = (ircMsgData *)calloc(1, sizeof(ircMsgData));

    if (*ptr != ':') {
        target = ptr;
        ptr = SeperateWord(target);
        strncpy(msgdata->target, target, CHANNELLEN);
    }

    if (*ptr == ':') {
        *ptr++ = '\0';
        strncpy(msgdata->data, ptr, 1024);
    }

    if (msgdata->target[0] == '\0' && msgdata->data[0] == '\0') {
        free(msgdata);
        return NULL;
    }

    return msgdata;
}

int ircSendPriv (ircClient *client, char *to, char *data, ...)
{
    char buf[1024];
    va_list val;
    ircsprintf(buf, 1023, data, val);
    return (ircSend(client, "PRIVMSG %s :%s", to, buf));
}

int ircSendNotice (ircClient *client, char *to, char *data, ...)
{
    char buf[1024];
    va_list val;
    ircsprintf(buf, 1023, data, val);
    return (ircSend(client, "NOTICE %s :%s", to, buf));
}

int ircSendQuit (ircClient *client, char *data, ...)
{
    char buf[1024];
    va_list val;
    ircsprintf(buf, 1023, data, val);
    return (ircSend(client, "QUIT :%s", buf));
}

int ircSendJoin (ircClient *client, char *to)
{
    return (ircSend(client,"JOIN %s", to));
}

int ircSendPart (ircClient *client, char *from, char *data, ...)
{
    char buf[1024];
    va_list val;
    ircsprintf(buf, 1023, data, val);
    return (ircSend(client, "PART %s :%s", from, buf));
}

int ircSendTopic (ircClient *client, char *to, char *data, ...)
{
    char buf[1024];
    va_list val;
    ircsprintf(buf, 1023, data, val);
    return (ircSend(client, "TOPIC %s :%s", to, buf));
}

int ircSendMode (ircClient *client, char *to, char *data, ...)
{
    char buf[1024];
    va_list val;
    ircsprintf(buf, 1023, data, val);
    return (ircSend(client, "MODE %s %s", to, buf));
}

int ircSendKick (ircClient *client, char *from, char *who, char *data, ...)
{
    char buf[1024];
    va_list val;
    ircsprintf(buf, 1023, data, val);
    return (ircSend(client, "KICK %s %s :%s", from, who, buf));
}
